﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Pixiview.Resources;
using Pixiview.UI;
using Pixiview.Utils;
using Xamarin.Essentials;
using Xamarin.Forms;

namespace Pixiview.Illust
{
    public partial class FavoritesPage : FavoriteIllustCollectionPage
    {
        private const int STEP = 20;

        private int startIndex;
        private int nextIndex;
        private bool flag = false;

        public FavoritesPage()
        {
            Resources.Add("cardView", GetCardViewTemplate());
            InitializeComponent();

            startIndex = -1;
            nextIndex = 0;
        }

        protected override bool IsFavoriteVisible => false;
        protected override ActivityIndicator LoadingIndicator => activityLoading;

        protected override void OnAppearing()
        {
            if (lastUpdated != LastUpdated)
            {
                startIndex = -1;
                nextIndex = 0;
                StartLoad();
            }
            else
            {
                var favorites = Stores.Favorites;
                if (favorites.Changed)
                {
                    favorites.Reload();
                    lastUpdated = default;
                    startIndex = -1;
                    nextIndex = 0;
                    StartLoad();
                }
            }
        }

        protected override IEnumerable<IllustItem> DoGetIllustList(IllustItem[] data)
        {
            return data;
        }

        protected override IllustItem[] DoLoadIllustData(bool force)
        {
            FavoriteList favs;
            if (startIndex < 0)
            {
                var favorites = Stores.GetFavoriteObject(flag);
                flag = false;
                if (favorites == null)
                {
                    return null;
                }
                favs = favorites.Illusts;
                startIndex = 0;
            }
            else
            {
                favs = Stores.Favorites;
            }
            var illusts = favs.Skip(startIndex).Take(STEP).ToArray();
            nextIndex = startIndex + STEP;
            if (illusts.Length == 0 || nextIndex >= Stores.Favorites.Count)
            {
                // reach the bottom
                startIndex = nextIndex;
            }
            return illusts;
        }

        private void FlowLayout_MaxHeightChanged(object sender, HeightEventArgs e)
        {
            SetOffset(e.ContentHeight - scrollView.Bounds.Height - SCROLL_OFFSET);
        }

        protected override bool CheckRefresh()
        {
            if (nextIndex > startIndex)
            {
                startIndex = nextIndex;
                return true;
            }
            return false;
        }

        private void ScrollView_Scrolled(object sender, ScrolledEventArgs e)
        {
            var y = e.ScrollY;
            OnScrolled(y);
        }

        public void Reload(bool force = false)
        {
            lastUpdated = default;
            if (force)
            {
                flag = true;
            }
            startIndex = -1;
            nextIndex = 0;
            StartLoad(force);
        }

        private async void Refresh_Clicked(object sender, EventArgs e)
        {
            if (IsLoading)
            {
                return;
            }
            await ScrollToTopAsync(scrollView);
            IsLoading = true;

            var offset = 16 - IndicatorMarginTop;
            activityLoading.Margin = new Thickness(0, loadingOffset - offset, 0, offset);
            activityLoading.Animate("margin", top =>
            {
                activityLoading.Margin = new Thickness(0, top, 0, offset);
            },
            loadingOffset - offset, 16 - offset, easing: Easing.CubicOut, finished: (v, r) =>
            {
                Task.Run(() =>
                {
                    var list = Stores.LoadOnlineFavorites();
                    if (list != null && list.Length > 0)
                    {
                        MainThread.BeginInvokeOnMainThread(() => ConfirmNext(list));
                    }
                    else
                    {
                        flag = false;
                        lastUpdated = default;
                        startIndex = -1;
                        nextIndex = 0;
                        MainThread.BeginInvokeOnMainThread(() => StartLoad(true));
                    }
                });
            });
        }

        private void CloseLoading(Action next = null)
        {
            var offset = 16 - IndicatorMarginTop;
            activityLoading.Animate("margin", top =>
            {
                activityLoading.Margin = new Thickness(0, top, 0, offset);
            },
            16 - offset, loadingOffset - offset, easing: Easing.CubicIn, finished: (v, r) =>
            {
                IsLoading = false;

                next?.Invoke();
            });
        }

        private async void ConfirmNext(IllustItem[] list)
        {
            var cancel = ResourceHelper.Cancel;
            var combine = ResourceHelper.FavoritesCombine;
            var replace = ResourceHelper.FavoritesReplace;
            var result = await DisplayActionSheet(
                ResourceHelper.FavoritesOperation,
                cancel,
                combine,
                replace);

            if (result == cancel)
            {
                return;
            }
            if (result == replace)
            {
                Stores.GetFavoriteObject().Illusts = new FavoriteList(list);
            }
            else if (result == combine)
            {
                // combine
                var nows = Stores.GetFavoriteObject().Illusts;
                // sync bookmarks from remote
                for (var i = nows.Count - 1; i >= 0; i--)
                {
                    var b = nows[i];
                    var bookmarkId = b.BookmarkId;
                    var bookmark = list.FirstOrDefault(f => f.Id == b.Id);
                    if (bookmark == null)
                    {
                        if (!string.IsNullOrEmpty(bookmarkId))
                        {
                            // not exists in remote any more
                            App.DebugPrint($"remove bookmark ({bookmarkId}) - {b.Id}: {b.Title}");
                            nows.RemoveAt(i);
                        }
                    }
                    else if (bookmarkId != bookmark.BookmarkId)
                    {
                        // update bookmark id
                        App.DebugPrint($"change bookmark ({bookmarkId}) to ({bookmark.BookmarkId}) - {b.Id}: {b.Title}");
                        b.BookmarkId = bookmark.BookmarkId;
                    }
                }
                // add bookmarks that exists in remote only
                var news = list.Where(f => !nows.Any(i => i.Id == f.Id)).ToArray();

                if (news.Length > 0)
                {
                    for (var i = 0; i < news.Length; i++)
                    {
                        var item = news[i];
                        App.DebugPrint($"add bookmark ({item.BookmarkId}) - {item.Id}: {item.Title}");
                        item.Image = StyleDefinition.DownloadBackground;
                    }
                    nows.InsertRange(0, news);
                }
                else
                {
                    CloseLoading();
                    return;
                }
            }
            else
            {
                return;
            }

            ParallelTask.Start(0, list.Length, Configs.MaxPageThreads, i =>
            {
                var illustItem = list[i];
                var data = Stores.LoadIllustPreloadData(illustItem.Id);
                if (data != null && data.illust.TryGetValue(illustItem.Id, out var illust))
                {
                    illust.CopyToItem(illustItem);
                    if (data.user.TryGetValue(illust.userId, out var user))
                    {
                        illustItem.ProfileUrl = user.image;
                    }
                    return true;
                }
                App.DebugError("load.favorite", $"cannot fetch preload data, {illustItem.Id}, disposed.");
                return false;
            }, () =>
            {
                Stores.SaveFavoritesIllusts();

                MainThread.BeginInvokeOnMainThread(() =>
                {
                    CloseLoading(() =>
                    {
                        flag = false;
                        lastUpdated = default;
                        StartLoad();
                    });
                });
            });
        }

        private async void ShareFavorites_Clicked(object sender, EventArgs e)
        {
            var file = Stores.FavoritesPath;
            await Share.RequestAsync(new ShareFileRequest
            {
                Title = ResourceHelper.Favorites,
                File = new ShareFile(file)
            });
        }
    }
}
